unit syntax_check_u;

{
funcs_all_sl    : all php- and project-function.   Format: func - filename
funcs_project_sl: all functions of project.        Format: func - filename
funcs_unused_sl : all unused functions in project. Format: func - filename
funcs_unknown_sl: all unknow functions.            Format: filename|y|x|func
func_calls_sl   : function-calls in one php-file.  Format: y|x|func
}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls,strutils;

type
  Tsyntax_check_f = class(TForm)
    main_p: TPanel;
    ok_b: TButton;
    break_b: TButton;
    lint_chb: TCheckBox;
    functions_unknown_chb: TCheckBox;
    functions_unused_chb: TCheckBox;
    variables_unknown_chb: TCheckBox;
    variables_unused_chb: TCheckBox;
    error_m: TMemo;
    top_p: TPanel;
    php_exe_l: TLabel;
    php_fn_e: TEdit;
    src_m: TMemo;
    procedure FormShow(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure ok_bClick(Sender: TObject);
    procedure break_bClick(Sender: TObject);
  private
    { Private-Deklarationen }
  public
    { Public-Deklarationen }
    rc:bool;
    procedure debug_lb_err_add(var s:string);
    function search_next(const s,search:string;start_c:integer):bool;

    procedure php_isolate(var s:string;fn:string;php_tag_ok:bool;var err:string);
    function php_isolate_fn(fn:string;var full_src:string;var err:string):string;

    procedure fn_funcs_sl_add(fn:string;sl:tstringlist);
    function word_prev(s:string;var start_c:integer):string;

    procedure syntax_check;
  end;

var
  syntax_check_f: Tsyntax_check_f;

implementation

uses main_u,syntax_check_php_lint_u,syntax_check_funcs_u,syntax_check_vars_u;

{$R *.dfm}

//------------------------------------------------------------------
//FORM
//------------------------------------------------------------------

procedure Tsyntax_check_f.FormShow(Sender: TObject);
begin
  rc:=false;
end;

procedure Tsyntax_check_f.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if      key=vk_return then ok_bclick(sender)
  else if key=vk_escape then break_bclick(sender);
end;

procedure Tsyntax_check_f.ok_bClick(Sender: TObject);
begin
  rc:=true;
  close;
end;

procedure Tsyntax_check_f.break_bClick(Sender: TObject);
begin
  close;
end;

//------------------------------------------------------------------
//SERVICE
//------------------------------------------------------------------

//copy error-strings (from php_isolate) to debug-listbox
procedure tsyntax_check_f.debug_lb_err_add(var s:string);
var
  c:integer;
  ss:string;
begin
  //nothing to do?
  s:=trim(s);if s='' then exit;
  repeat
    //search end of line
    c:=pos(_cr,s);if c=0 then c:=length(s)+1;

    //isolate line
    ss:=trim(copy(s,1,c-1));
    if ss<>'' then
    begin
      //add line to debug-listnox
      if main_f.debug_lb.Items.indexof(ss)=-1 then main_f.debug_lb.Items.add(ss);
    end;
    s:=trim(copy(s,c+1,length(s)));
  until s='';
  s:='';
end;

//find next 'search' in 's' from position 'start_c'
function tsyntax_check_f.search_next(const s,search:string;start_c:integer):bool;
var
  l,c:integer;
begin
  result:=false;l:=length(search);if start_c+l>length(s) then exit;
  for c:=1 to l do
  begin
    if ansilowercase(s[start_c+c-1])<>search[c] then exit;
  end;
  result:=true;
end;

//add all functions-headers of php-file 'fn' to 'funcs_project_sl'
procedure tsyntax_check_f.fn_funcs_sl_add(fn:string;sl:tstringlist);
var
  c,cstart:integer;
  s,su,func,err:string;
begin
  //nothing to do?
  if not fileexists(fn) then exit;

  //load source into src_m-memo, copy it to 's'
  try
    src_m.Lines.LoadFromFile(fn);
  except
    exit;
  end;
  s:=src_m.text;
  src_m.clear;

  //isolate php-parts from 'su'
  su:=ansilowercase(s);
  php_isolate(su,fn,true,err);if err<>'' then debug_lb_err_add(err);

  //scan for functions
  c:=1;
  repeat
    //function-header found? function tst($a)
    c:=posex('function ',su,c+1);if c=0 then break;
    cstart:=c;

    //scan end of function-header
    c:=posex(')',su,cstart);if c=0 then break;

    //delete 'function'
    func:=trim(copy(s,cstart+9,c-cstart-9+1));

    //add function header to sl
    sl.add(func+' - '+main_f.short_fn(fn));

    inc(c);
  until c=0;
end;

//search previous word in 's' from 'start_c'
//function test($a,$b -> $b
function tsyntax_check_f.word_prev(s:string;var start_c:integer):string;
var
  word:string;
  ch:char;
begin
  word:='';
  repeat
    dec(start_c);
    ch:=s[start_c];
    if ch in [':','=',';',',','.','-','+','/','*','(',')','!','@','<','>','{','}','[','?'] then break;
    word:=trim(ch+word);
  until
    ((trim(word)<>'') and((ch=' ')or(ch=#13)or(ch=#10)))or
    (start_c=0);
  result:=word;
end;

//------------------------------------------------------------------
//PHP-ISOLATE
//------------------------------------------------------------------

//isolate all php-parts from 's'
//if php_tag_ok then check for php-tags '<?'
//<html><?php echo 'test';?></html> => echo '....';
procedure tsyntax_check_f.php_isolate(var s:string;fn:string;php_tag_ok:bool;var err:string);

  procedure err_tag_open(c:integer;title,tag:string);
  var
    x,y:integer;
  begin
    if c=0 then exit;
    main_f.pos_to_xy(s,c,x,y);
    err:=err+_db_error+'|'+fn+'|'+main_f.itoa(y)+'|'+main_f.itoa(x)+'|||'+title+' not closed ('+tag+' missing)'+_cr;
  end;

  procedure warning_tag_open(c:integer;title,tag:string);
  var
    x,y:integer;
  begin
    if c=0 then exit;
    main_f.pos_to_xy(s,c,x,y);
    err:=err+_db_warning+'|'+fn+'|'+main_f.itoa(y)+'|'+main_f.itoa(x)+'|||'+title+' open inside '+title+'? '+tag+' missing!'+_cr;
  end;

  procedure warning_tag_close(c:integer;title,tag:string);
  var
    x,y:integer;
  begin
    if c=0 then exit;
    main_f.pos_to_xy(s,c,x,y);
    err:=err+_db_warning+'|'+fn+'|'+main_f.itoa(y)+'|'+main_f.itoa(x)+'|||'+title+' closed, but '+tag+' missing!'+_cr;
  end;

const
  _replace_ch='.';
var
  replace_ch:char;
  c:integer;
  cast_ok,quote_ok,apostrophe_ok,comment_block_ok,comment_line_ok,php_ok:bool;
  php_c,quote_c,comment_block_c,apostrophe_c:integer;
  php_xml_ok,var_ok:bool;
begin
  replace_ch:='.';
  fn:=main_f.short_fn(fn);

  //set counter to zero
  php_c:=0;quote_c:=0;comment_block_c:=0;apostrophe_c:=0;

  //set flags to false
  php_ok:=false;if not php_tag_ok then php_ok:=true;
  php_xml_ok:=false;
  cast_ok:=false;
  comment_block_ok:=false;comment_line_ok:=false;
  quote_ok:=false;apostrophe_ok:=false;var_ok:=false;

  //check all chars of s
  for c:=1 to length(s) do
  begin

    //is cast-mode active?
    if cast_ok then
    begin
      //yes, deactivate it and replace next char
      cast_ok:=false;
      s[c]:=replace_ch;
      continue;
    end;

    //is the a space-char?
    if s[c]=' ' then
    begin
      //yes: deaktivate var-mode
      var_ok:=false;
      continue;
    end;

    //is there a carriage or return-char
    if(s[c]=#13)or(s[c]=#10)then
    begin
      //yes: deactivate var- and line-comment-mode
      var_ok:=false;
      comment_line_ok:=false;
      continue;
    end;

    //php-tags needed?
    if php_tag_ok then
    begin
      //yes: if in string- or comment-mode then ognore php-tags
      if not quote_ok and not apostrophe_ok and not comment_block_ok then
      begin
        //check for php-open-tag, ignore xml-open-tag
        if php_ok and(search_next(s,'<?',c))and not(search_next(s,'<?xml',c))then
        begin
          //doubled php-open-tag
          warning_tag_open(c,'PHP-Tag','?>');
        end;

        //check for php-close-tag
        if not php_ok and(search_next(s,'?>',c))then
        begin
          //are you in xml-mode?
          if not php_xml_ok then warning_tag_close(c,'PHP-Tag','<?')
                            else php_xml_ok:=false;
        end;

        //check for guilty php-open-tag
        if not php_ok and(search_next(s,'<?',c))then
        begin
          //is it a xml-open-ttag?
          if (search_next(s,'<?xml',c))then php_xml_ok:=true
          else
          begin
            //ok, start php-mode
            php_ok:=true;
            php_c:=c;
          end;
        end;
        //check for php-close-tag
        if php_ok and (search_next(s,'?>',c)) then
        begin
          //ok, close php-mode
          php_ok:=false;
          php_c:=0;
        end;
      end;
      //php-mode active?
      if not php_ok then
      begin
        //no: replace char
        s[c]:=replace_ch;
        continue;
      end;
    end;

    //check for comment-block
    if not quote_ok and not apostrophe_ok and not comment_line_ok then
    begin
      //check for comment-block-close
      if not comment_block_ok and (search_next(s,'*/',c)) then
        warning_tag_close(c,'Comment-Block','/*');

      //check for commnt-block-open
      if not comment_block_ok and(search_next(s,'/*',c)) then
      begin
        //comment-block-mode active
        comment_block_ok:=true;
        comment_block_c:=c;
      end;
      //check for guilty comment-block-close
      if comment_block_ok and (search_next(s,'*/',c)) then
      begin
        //comment-block-mode inactive
        comment_block_ok:=false;
        comment_block_c:=0;
      end;
    end;
    //comment-block active?
    if comment_block_ok then
    begin
      //yes: replace char
      s[c]:=replace_ch;
      continue;
    end;

    //check for comment-line
    if not quote_ok and not apostrophe_ok and not comment_block_ok then
    begin
      //guilty comment-line-open found?
      if not comment_line_ok and(search_next(s,'//',c)) then comment_line_ok:=true;
    end;
    //comment-line-mode active?
    if comment_line_ok then
    begin
      //yes: ignore char
      s[c]:=replace_ch;
      continue;
    end;

    //chaeck fpr php-variable
    if not var_ok and(s[c]='$') then var_ok:=true
    else if not(s[c] in ['a'..'z','A'..'Z','0'..'9','_']) then var_ok:=false;

    //check for quote-mode
    if not apostrophe_ok then
    begin
      if s[c]='"' then
      begin
        quote_ok:=not quote_ok;
        if quote_ok then quote_c:=c else quote_c:=0;
      end;
    end;
    if quote_ok then
    begin
      //save begin of cast-mode?
      cast_ok:=(s[c]='\');
      if not var_ok then s[c]:=replace_ch;
      continue;
    end;

    //check for apostrophe-mode
    if s[c]='''' then
    begin
      apostrophe_ok:=not apostrophe_ok;
      if apostrophe_ok then apostrophe_c:=c else apostrophe_c:=0;
    end;
    if apostrophe_ok then
    begin
      //save begin of cast-mode?
      cast_ok:=(s[c]='\');
      var_ok:=false;
      s[c]:=replace_ch;
      continue;
    end;
  end;

  //are there any open/close-tag-errors?
  err_tag_open(comment_block_c,'Comment-Block','/*');
  err_tag_open(quote_c,'Quote-String','"');
  err_tag_open(apostrophe_c,'Apostrophe-String','''');
  err_tag_open(php_c,'PHP-Tag','?>');
end;

//isolate php-parts from source from php-file 'fn'
//'src' is complete source
//'result' is isolated php-source
function tsyntax_check_f.php_isolate_fn(fn:string;var full_src:string;var err:string):string;
begin
  //nothing to do?
  if not fileexists(fn) then exit;

  //get php-file-source
  src_m.Lines.loadfromfile(fn);
  full_src:=src_m.text;
  src_m.clear;
  result:=full_src;

  //isolate php-parts in 'result'
  php_isolate(result,fn,true,err);if err<>'' then debug_lb_err_add(err);
end;

//------------------------------------------------------------------
//SYNTAX CHECK
//------------------------------------------------------------------

//check syntax of all project files
//1) Syntax-Check per PHP-Lint
//2) Find unused and unknown Functions
//3) Find unused and unknown Variables
procedure tsyntax_check_f.syntax_check;
var
  funcs_all_sl,funcs_project_sl,
  funcs_unused_sl,funcs_unknown_sl,
  func_calls_sl:tstringlist;
  fn,long_fn,php_src,full_src,err,func_name:string;
  r,x,y:integer;
  prj_c:integer;
begin
  screen.cursor:=crhourglass;
  funcs_all_sl:=tstringlist.Create;
  funcs_project_sl:=tstringlist.Create;
  funcs_unused_sl:=tstringlist.Create;
  funcs_unknown_sl:=tstringlist.Create;
  func_calls_sl:=tstringlist.Create;
  try
    //save project an editor
    main_f.debug_lb.clear;
    main_f.src_save;
    main_f.project_save(main_f.project_fn);

    //scan all functions of all project files
    screen.cursor:=crhourglass;
    prj_c:=main_f.project_lb.items.count;
    for r:=0 to prj_c-1 do
    begin
      fn:=main_f.project_lb.items[r];
      long_fn:=main_f.long_fn(main_f.project_lb.items[r]);
      main_f.status_set('Syntax-Check: ['+inttostr(r+1)+'/'+inttostr(prj_c)+'] Scan functions of "'+long_fn+'" ...');
      application.ProcessMessages;
      //scan and add all functions-headers to funcs_project_sl
      fn_funcs_sl_add(long_fn,funcs_project_sl);
    end;

    //assign result to project-functions-listbox
    main_f.project_functions_lb.items.assign(funcs_project_sl);
    main_f.lb_scrollwidth_set(main_f.project_functions_lb);

    //assign project-functions to unused-function-listbox
    //kill later found used functions from this stringlist
    funcs_unused_sl.assign(funcs_project_sl);

    //all possible functions: add php-functions to project-functions
    funcs_all_sl.assign(funcs_project_sl);
    for r:=0 to main_f.php_functions_lb.items.count-1 do
    begin
      funcs_all_sl.add(main_f.php_functions_lb.items[r]+' - PHP');
    end;

    //check syntax, functions and variables of all project-files
    for r:=0 to prj_c-1 do
    begin
      screen.cursor:=crhourglass;
      fn:=main_f.project_lb.items[r];
      long_fn:=main_f.long_fn(main_f.project_lb.items[r]);
      main_f.status_set('Syntax-Check: ['+inttostr(r+1)+'/'+inttostr(prj_c)+'] Check "'+long_fn+'" ...');
      application.ProcessMessages;

      //php-lint-syntax-check
      if lint_chb.Checked then syntax_check_php_lint_u.php_lint(long_fn);

      //no more to do?
      if not(
        functions_unknown_chb.checked or
        functions_unused_chb.checked or
        variables_unknown_chb.checked or
        variables_unused_chb.checked
      ) then continue;

      //isolate php-parts from source
      php_src:=php_isolate_fn(long_fn,full_src,err);

      //check functions
      func_calls_sl.clear;
      syntax_check_funcs_u.check(
        fn,full_src,php_src,
        funcs_all_sl,
        funcs_project_sl,
        funcs_unused_sl,
        funcs_unknown_sl,
        func_calls_sl
      );

      //check variables
      syntax_check_vars_u.check(fn,full_src,php_src,funcs_all_sl,func_calls_sl);
    end;

    //add unknown functions to debug-listbox
    if functions_unknown_chb.Checked then
    begin
      for r:=0 to funcs_unknown_sl.Count-1 do
      begin
        fn:=trim(main_f.delim_param_get(funcs_unknown_sl[r],0,'|','0'));
        y:=strtoint(trim(main_f.delim_param_get(funcs_unknown_sl[r],1,'|','0')));
        x:=strtoint(trim(main_f.delim_param_get(funcs_unknown_sl[r],2,'|','0')));
        func_name:=trim(main_f.delim_param_get(funcs_unknown_sl[r],3,'|',''));
        main_f.debug_add(_db_warning,fn,y,x,func_name,'','UNKNOWN function?');
      end;
    end;

    //add unused functions to debug-listbox
    if functions_unused_chb.Checked then
    begin
      for r:=0 to funcs_unused_sl.Count-1 do
      begin
        func_name:=trim(main_f.delim_param_get(funcs_unused_sl[r],0,'-',''));
        fn:=trim(main_f.delim_param_get(funcs_unused_sl[r],1,'-',''));
        main_f.debug_add(_db_warning,fn,0,0,func_name,'','UNUSED function?');
      end;
    end;
    main_f.status_set('Syntax-Check: ['+inttostr(prj_c)+'/'+inttostr(prj_c)+'] Ok, all files of project checked.');
  finally
    func_calls_sl.free;
    funcs_all_sl.Free;
    funcs_unknown_sl.Free;
    funcs_unused_sl.Free;
    funcs_project_sl.Free;
    main_f.lb_scrollwidth_set(main_f.debug_lb);
    screen.cursor:=crdefault;
  end;
end;

end.

