unit syntax_check_vars_u;

interface

uses windows,sysutils,Classes,strutils;

procedure php_isolate_root(var php_s:string);
procedure check(fn,src,php_s:string;funcs_all_sl,func_calls_sl:tstringlist);

implementation

uses main_u,syntax_check_u;

//--------------------------------------------------
//isolate root-source (function-parts hided)
procedure php_isolate_root(var php_s:string);
const
  _replace_ch='.';
var
  c,brace_c:integer;
  func_ok:bool;
begin
  brace_c:=0;func_ok:=false;
  for c:=1 to length(php_s) do
  begin
    //ignore some chars
    if(php_s[c]=#13)or(php_s[c]=#10)or(php_s[c]=' ')then continue;

    //function-start found? Then function-mode active
    if not func_ok and(syntax_check_f.search_next(php_s,'function ',c))then func_ok:=true;

    //class-start found? Then function-mode active
    if not func_ok and(syntax_check_f.search_next(php_s,'class ',c))then func_ok:=true;

    //function-body-start found?
    if func_ok and(php_s[c]='{')then inc(brace_c);

    //function-end reached?
    if func_ok and(php_s[c]='}')then
    begin
      dec(brace_c);
      //function-end reached? Then function-mode off
      if brace_c=0 then func_ok:=false;
    end;

    //replace all chars inside a function
    if func_ok then php_s[c]:=_replace_ch;
  end;
end;

//------------------------------------------------------------
procedure check(fn,src,php_s:string;funcs_all_sl,func_calls_sl:tstringlist);
var
  vars_funcs_sl,               //all variables of a function. Format: src-pos|var
  vars_unused_sl:tstringlist;  //unused vars of a function.   Format: src-pos|var
  current_func_name:string;

  //---------------------------------------------------------
  procedure sl_add(sl:tstringlist;c:integer;v:string);
  var
    s:string;
  begin
    s:=inttostr(c);while length(s)<10 do s:='0'+s;
    sl.add(s+'|'+v);
  end;

  //---------------------------------------------------------
  procedure var_funcs_sl_add(c:integer;v:string);
  begin
    sl_add(vars_funcs_sl,c,v);
  end;

  //---------------------------------------------------------
  procedure unused_sl_add(c:integer;v:string);
  begin
    sl_add(vars_unused_sl,c,v);
  end;

  //------------------------------------------------------------
  //put all variables of comma-string to var_funcs_sl
  //$a,$b='xxx',$$c -> Variables: $a,$b,$c
  procedure var_funcs_sl_comma_str_add(comma_str:string;found_c:integer);
  var
    c,cc:integer;
    v:string;
  begin
    //scann all variables in comma-string
    repeat
      //search comma
      c:=posex(',',comma_str);

      //no comma found? Then reached end of line
      if c=0 then c:=length(comma_str)+1;

      //isolate variable
      v:=copy(comma_str,1,c-1);

      //clean up variable: $a='..' -> $a
      cc:=pos('=',v);if cc>0 then v:=copy(v,1,cc-1);

      //clean up var-variable: &$a -> $a
      v:=stringreplace(v,'&','',[rfreplaceall]);
      v:=trim(v);

      //add variable to variable-list with position in source
      var_funcs_sl_add(found_c,v);

      //prepare for next variable in string
      comma_str:=copy(comma_str,c+1,length(comma_str));
    until comma_str='';
  end;

  //---------------------------------------------------------
  //add function parameter to function-variable-list
  function var_funcs_sl_param_add(start_c:integer):integer;
  var
    s_c,e_c:integer;
    php_param:string;
  begin
    result:=start_c;

    //isolate parameter: function tst($a,$b,$c)
    s_c:=posex('(',php_s,start_c)+1;
    e_c:=posex(')',php_s,s_c);

    //cut funktion-name for debug-list
    //function tst
    current_func_name:=trim(copy(php_s,start_c,s_c-start_c-1));
    //tst
    current_func_name:=trim(copy(current_func_name,length('function')+1,length(current_func_name)));

    //isolate paramters: $a,$b,$c
    php_param:=copy(php_s,s_c,e_c-s_c);

    //no parameters? Then exit
    if php_param='' then exit;

    //put all variables of comma-paramter-string to var_funcs_sl
    var_funcs_sl_comma_str_add(php_param,e_c);

    //return pointer if begin of function-body
    result:=e_c;
  end;

  //------------------------------------------------------------
  //scan for globals in source-part
  procedure var_funcs_sl_globals_add(start_c,end_c:integer);
  const
    _global='global ';
  var
    s:string;
    s_c,e_c:integer;
  begin
    s_c:=0;
    repeat
      //search for something like: global $a,$b;
      s_c:=posex(_global,php_s,s_c+1);if(s_c=0)or(s_c>end_c) then exit;

      //search end of global-line
      e_c:=posex(';',php_s,s_c+1);if(e_c=0)or(e_c>end_c) then exit;

      //isolate the variables: $a,$b
      s:=trim(copy(php_s,s_c+length(_global),e_c-s_c-length(_global)));

      //put all variables of globals-comma-string to var_funcs_sl
      var_funcs_sl_comma_str_add(s,s_c);
    until s_c=0;
  end;

  //------------------------------------------------------------
  //scan and add the reference-paramters of function calls
  procedure var_funcs_sl_ref_param_add(start_c,end_c:integer);
  var
    sl:tstringlist;
    r,rr,c,x,y,yy,min_y,max_y:integer;
    s,err,
    def_s,call_s,
    def_name,call_name,
    def_params,call_params,
    def_param,call_param:string;
  begin
    sl:=tstringlist.Create;
    try
      //get start-line of source-part (function)
      main_f.pos_to_xy(php_s,start_c,x,min_y);

      //get end-line of source-part (function)
      main_f.pos_to_xy(php_s,end_c,x,max_y);

      //check all function-calls
      for r:=0 to func_calls_sl.count-1 do
      begin
        //get entry: y|x|tst($a,$b)
        s:=func_calls_sl[r];
        try
          //get line of function call
          y:=strtoint(main_f.delim_param_get(s,0,'|',''));

          //is ist part of the corrent source-part (function)?
          if(y<min_y)or(y>max_y) then continue;

          //yes, get function-header: tst($a,$b)
          call_s:=main_f.delim_param_get(s,2,'|','');

          //isolate function-name: tst
          c:=pos('(',call_s);if c=0 then continue;
          call_name:=trim(copy(call_s,1,c-1));

          //isolate paramter: $a,$b)
          call_params:=trim(copy(call_s,c+1,length(call_s)));

          //cut brace at end: $a,$b
          call_params:=trim(copy(call_params,1,length(call_params)-1));

          //isolate php-source
          //$a,'test','test2',$b => $a,'....','.....',$b
          syntax_check_f.php_isolate(call_params,'',false,err);


          //search the function-definition in project
          for rr:=0 to funcs_all_sl.count-1 do
          begin
            //get entry: tst($x,&$y) - fn.php
            def_s:=funcs_all_sl[rr];

            //isolate function-name
            c:=pos('(',def_s);if c=0 then continue;
            def_name:=trim(copy(def_s,1,c-1));

            //found definition to call?
            if call_name<>def_name then continue;

            //yes, isolate parameter: $a,$b
            def_params:=trim(copy(def_s,c+1,length(def_s)));
            def_params:=trim(copy(def_params,1,pos(')',def_params)-1));

            //scan for reference-parameter
            c:=0;
            repeat
              //get parameter number 'c'
              call_param:=trim(main_f.delim_param_get(call_params,c,',',''));
              def_param:=trim(main_f.delim_param_get(def_params,c,',',''));

              //parameter exists in definition and call?
              if(call_param<>'')and(def_param<>'')then
              begin
                //yes: check if it a reference-parameter
                if def_param[1]='&' then
                begin
                  //yes: get line-position of the reference-parameter
                  yy:=main_f.xy_to_pos(php_s,0,y-1);
                  //add reference-parameter to 'var_funcs_sl'
                  var_funcs_sl_add(yy,call_param);
                end;
              end;
              inc(c);
            until call_param='';
            break;
          end;
        except
        end;
      end;
    finally
      sl.free;
    end;
  end;

  //---------------------------------------------------------
  //check if variable 'v' is defined in 'var_funcs_sl'
  function var_in_var_funcs_sl(v:string):integer;
  var
    r,c:integer;
    s:string;
  begin
    result:=0;
    for r:=0 to vars_funcs_sl.count-1 do
    begin
      s:=vars_funcs_sl[r];
      c:=pos('|',s);
      if copy(s,c+1,length(s))<>v then continue;
      result:=strtoint(copy(s,1,c-1));
      exit;
    end;
  end;

  //------------------------------------------------------------
  procedure unused_sl_del(s:string);
  var
    r:integer;
    v:string;
  begin
    //verwendung, also aus unused_sl raus, falls vorhanden
    for r:=vars_unused_sl.count-1 downto 0 do
    begin
      v:=main_f.delim_param_get(vars_unused_sl[r],1,'|','');
      if v<>s then continue;
      vars_unused_sl.Delete(r);
      exit;
    end;
  end;

  //------------------------------------------------------------
  //isolate varibale-name in source-part
  function var_isolate(php:string;start_c,max_c:integer;var end_c:integer):string;
  var
    def_c,c,s_c:integer;
    s,ss:string;
  begin
    //fill a guilty variable-string
    s:='$';
    for c:=start_c+1 to max_c do
    begin
      if not(php[c] in ['a'..'z','A'..'Z','0'..'9','_']) then
      begin
        end_c:=c;
        break;
      end;
      s:=s+php[c];
    end;

    //variable empty=
    if s='$' then
    begin
      //oops, this should never happen
      result:='';
      exit;
    end;

    //is it a 'this'-variable like: $this->tst="x";
    if s='$this' then
    begin
      //yes: Ignore it, because is defined in a class
      result:='';
      exit;
    end;

    //is it an exception like: catch(exception $e)
    s_c:=end_c-length(s);
    ss:=syntax_check_f.word_prev(php,s_c);
    if lowercase(ss)='exception' then
    begin
      //ignore it beauce is defined by system
      result:='';
      exit;
    end;

    //variable is already defined?
    def_c:=var_in_var_funcs_sl(s);
    if(def_c>0)and(def_c<end_c) then
    begin
      //yes, the use auf this variable is ok
      result:='';

      //delete variable from unused-variable-list
      unused_sl_del(s);
      exit;
    end;

    //check if variable is defined here: $x='test';
    for c:=end_c to max_c-1 do
    begin
      if not(php[c] in [#9,#10,#13,' ','=']) then break;
      if (php[c]<>'=') then continue;

      //is it a comparemnt? if($a=='test')
      if php[c+1]='=' then break;

      //okay, variable is obviously defined here
      var_funcs_sl_add(end_c,s);

      //add variable to unused-list because definition is not use
      unused_sl_add(start_c,s);

      result:='';
      exit;
    end;

    //variable is not defined but used
    result:=s;

    //delete variable from unused-variable-list
    unused_sl_del(s);
  end;

  //------------------------------------------------------------
  //check source-part 'start_c' to 'end_c' for variables
  procedure func_vars_check(start_c,end_c:integer;func_ok:bool);
  var
    c,body_c,def_c,e_c,x,y,r:integer;
    v,s,msg:string;
  begin
    vars_funcs_sl.clear;
    vars_unused_sl.clear;

    //define superglobals
    var_funcs_sl_add(1,'$GLOBALS');
    var_funcs_sl_add(1,'$_SERVER');
    var_funcs_sl_add(1,'$_GET');
    var_funcs_sl_add(1,'$_POST');
    var_funcs_sl_add(1,'$_FILES');
    var_funcs_sl_add(1,'$_COOKIE');
    var_funcs_sl_add(1,'$_SESSION');
    var_funcs_sl_add(1,'$_REQUEST');
    var_funcs_sl_add(1,'$_ENV');

    body_c:=start_c;current_func_name:='';

    //function-mode? Then get function-parameters
    if func_ok then body_c:=var_funcs_sl_param_add(start_c);

    //scan and add the globals in source-part
    var_funcs_sl_globals_add(body_c,end_c);

    //scan and add the reference-paramter of function calls
    var_funcs_sl_ref_param_add(body_c,end_c);

    //search for all variables in source-part (function)
    //check if they are defined in 'var_funcs_sl' or in source
    c:=body_c;
    repeat
      //find next variable-start
      c:=posex('$',php_s,c+1);if(c=0)or(c>end_c) then break;

      //isolate variable-name
      v:=var_isolate(php_s,c,end_c,e_c);

      //check for unknown variables?
      if syntax_check_f.variables_unknown_chb.checked  and(v<>'') then
      begin
        msg:='';

        //yes: check if variable is defined
        def_c:=var_in_var_funcs_sl(v);

        //variable not defined?
        if def_c=0 then msg:='UNKNOWN variable?';

        //variable defined but after use-position?
        if def_c>e_c then msg:='UNKNOWN variable (first DEFINED after use)?';

        if msg<>'' then
        begin
          //variable-use-error
          //get line-position
          main_f.pos_to_xy(src,c-1,x,y);
          //add error to debug-listbox
          main_f.debug_add(_db_warning,fn,y,x,current_func_name,v,msg);
        end;
      end;
      c:=e_c;
    until c=0;

    //check for unused variables?
    if not syntax_check_f.variables_unused_chb.checked then exit;

    //get all found unused variables
    for r:=0 to vars_unused_sl.count-1 do
    begin
      //get entry: '0000000565|$hurz'
      s:=vars_unused_sl[r];

      //get source-position
      c:=strtoint(main_f.delim_param_get(s,0,'|','1'));

      //transform src-pos to x/y-coordinates
      //main_f.pos_to_xy(src,c-length(v)-1,x,y);
      main_f.pos_to_xy(src,c-1,x,y);

      //get variable-name
      v:=main_f.delim_param_get(s,1,'|','1');

      //add unused variable to debug-list
      main_f.debug_add(_db_warning,fn,y,x,current_func_name,v,'UNUSED variable?');
    end;
  end;

var
  c,start_c:integer;
  func_ok:bool;
  brace_c:integer;
begin
  //nothing to do?
  if not(syntax_check_f.variables_unknown_chb.checked or syntax_check_f.variables_unused_chb.checked)then exit;

  vars_funcs_sl:=tstringlist.Create;
  vars_unused_sl:=tstringlist.Create;
  try
    //need for chronolgy of uese
    vars_funcs_sl.sorted:=true;

    //scan php-source for variables
    func_ok:=false;brace_c:=0;start_c:=0;
    for c:=1 to length(php_s) do
    begin
      //find function-header?
      if not func_ok and(syntax_check_f.search_next(php_s,'function ',c))then
      begin
        //function-mode active
        func_ok:=true;
        //save function-start
        start_c:=c;
      end;
      //grow brace-counter?
      if func_ok and(php_s[c]='{')then inc(brace_c);

      //function-mode and close-brace?
      if func_ok and(php_s[c]='}')then
      begin
        dec(brace_c);
        //function-end reached?
        if brace_c=0 then
        begin
          //ok, function-end reached
          func_ok:=false;

          //check source-part for variables
          func_vars_check(start_c,c,true);
        end;
      end;
    end;

    //check now all root-vars

    //isolate root-source (function-parts hided)
    php_isolate_root(php_s);

    //check source-part for variables
    func_vars_check(1,length(php_s),false);
  finally
    vars_unused_sl.free;
    vars_funcs_sl.Free;
  end;
end;

end.

